# Copyright 2021 Adobe. All rights reserved.

"""
Storage for intermediate state built by the hinter object
"""

import weakref
import logging
import bisect
from enum import IntEnum


from . import Number
from .glyphData import feq, pathElement, stem
from typing import (
    Any,
    Callable,
    Sequence,
    Set,
    Protocol,
)
import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

log: logging.Logger = logging.getLogger(__name__)

# instanceStemState stores a hinter as one of its members. We'd
# love to be able to use the abstract base class `dimensionHinter`
# as the type for that member, but that would mean this module
# imports hinter which in turn imports this module. Instead, we
# type check against a "protocol" class (a "trait" in other languages)
# declaring only the method that instanceStemState will expect to
# find on the object. `...` stands for "we don't care, nobody's
# calling it anyway".


class AHinter(Protocol):
    options: Any

    def inBand(self, loc, isBottom: bool = False) -> bool:
        ...


class hintSegment:
    """Represents a hint "segment" (one side of a potential stem)"""

    class sType(IntEnum):
        LINE = 1
        BEND = 2
        CURVE = 3
        LSBBOX = 4  # Calculated from the lower bound of a segment
        USBBOX = 5  # Calculated from the upper bound of a segment
        LGBBOX = 6  # Calculated from the lower bound of the whole glyph
        UGBBOX = 7  # Calcualted from the upper bound of the whole glyph
        GHOST = 8

    # This is the *ideal* type hint, but 3.8 doesn't support it.
    # pe: weakref.ReferenceType[pathElement | None]
    pe: Callable[[], pathElement] | None

    def __init__(self, aloc: float, oMin: float, oMax: float,
                 pe: pathElement | None, typ, bonus, isV,
                 isInc, desc: str) -> None:
        """
        Initializes the object

        self.loc is the segment location in the aligned dimension
        (x for horizontal, y for vertical)

        self.min and self.max are the extent of the segment in the
        opposite dimension

        self.bonus is 0 for normal segments and more for "special"
        segments (e.g. those within a BlueValue)

        self.type indicates the source of the segment (with GHOST
        being specially important)

        self.isV records the dimension (False for horizontal, etc.)

        self.desc is a string with a more detailed indication of
        how the segment was derived)

        self.pe is a weak reference to the pathElement (spline) from
        which the segment was derived

        self.hintval, self.replacedBy, and self.deleted are state used
        by the hinter
        """
        self.loc = aloc
        self.min = oMin
        self.max = oMax
        self.bonus = bonus
        self.type = typ
        self.isV = isV
        self.isInc = isInc
        self.desc = desc
        if pe:
            self.pe = weakref.ref(pe)  # type: ignore
        else:
            self.pe = None
        self.hintval: stemValue | None = None
        self.replacedBy: Self | None = None
        self.deleted = False
        self.suppressed = False

    def __eq__(self, other) -> bool:
        return feq(self.loc, other.loc)

    def __lt__(self, other) -> bool:
        return self.loc < other.loc

    def isBend(self) -> bool:
        return self.type == self.sType.BEND

    def isCurve(self) -> bool:
        return self.type == self.sType.CURVE

    def isLine(self) -> bool:
        return self.type == self.sType.LINE or self.isBBox()

    def isBBox(self) -> bool:
        return self.type in (self.sType.LSBBOX, self.sType.USBBOX,
                             self.sType.LGBBOX, self.sType.UGBBOX)

    def isGBBox(self) -> bool:
        return self.type in (self.sType.LGBBOX, self.sType.UGBBOX)

    def isUBBox(self) -> bool:
        return self.type in (self.sType.USBBOX, self.sType.UGBBOX)

    def isGhost(self) -> bool:
        return self.type == self.sType.GHOST

    def current(self, orig=None) -> Self:
        """
        Returns the object corresponding to this object relative to
        self.replacedBy
        """
        if self.replacedBy is None:
            return self
        if self is orig:
            log.error("Cycle in hint segment replacement")
            return self
        if orig is None:
            orig = self
        return self.replacedBy.current(orig=self)

    def show(self, label: str) -> None:
        """Logs a debug message about the segment"""
        if self.isV:
            pp = (label, 'v', self.loc, self.min, self.loc, self.max,
                  self.desc)
        else:
            pp = (label, 'h', self.min, self.loc, self.max, self.loc,
                  self.desc)
        log.debug("%s %sseg %g %g to %g %g %s" % pp)


HintSegListWithType = tuple[str, list[hintSegment]]


class stemValue:
    """Represents a potential hint stem"""

    def __init__(
        self,
        lloc: Number,
        uloc: Number,
        val: Number,
        spc,
        lseg: hintSegment,
        useg: hintSegment,
        isGhost: bool = False,
    ) -> None:
        assert lloc <= uloc
        self.val = val
        self.spc = spc
        self.lloc = lloc
        self.uloc = uloc
        self.isGhost = isGhost
        self.pruned = False
        self.merge = False
        self.lseg = lseg
        self.useg = useg
        self.best: stemValue | None = None
        self.initialVal = val
        self.idx: int | None = None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, stemValue):
            return False
        slloc, suloc = self.ghosted()
        olloc, ouloc = other.ghosted()
        return slloc == olloc and suloc == ouloc

    def __lt__(self, other: Self) -> bool:
        """Orders values by lower and then upper location"""
        slloc, suloc = self.ghosted()
        olloc, ouloc = other.ghosted()
        return (slloc < olloc or (slloc == olloc and suloc < ouloc))

    # c.f. page 22 of Adobe TN #5177 "The Type 2 Charstring Format"
    def ghosted(self) -> tuple[Number, Number]:
        """Return the stem range but with ghost stems normalized"""
        lloc, uloc = self.lloc, self.uloc
        if self.isGhost:
            if self.lseg.type == hintSegment.sType.GHOST:
                lloc = uloc
                uloc = lloc - 20
            else:
                uloc = lloc
                lloc = uloc + 21
        return (lloc, uloc)

    def compVal(
        self,
        spcFactor: int | float = 1,
        ghostFactor: int | float = 1
    ) -> tuple[Any, Any]:
        """Represent self.val and self.spc as a comparable 2-tuple"""
        v = self.val
        if self.isGhost:
            v *= ghostFactor
        if self.spc > 0:
            v *= spcFactor
        return (v, self.initialVal)

    def show(self, isV, typ) -> None:
        """Add a log message with the content of the object"""
        tags = ('v', 'l', 'r', 'b', 't') if isV else ('h', 'b', 't', 'l', 'r')
        start = "%s %sval %s %g %s %g v %g s %g %s" % (typ, tags[0], tags[1],
                                                       self.lloc, tags[2],
                                                       self.uloc, self.val,
                                                       self.spc,
                                                       'G' if self.isGhost
                                                       else '')
        if self.lseg is None:
            log.debug(start)
            return
        log.debug("%s %s1 %g %s1 %g  %s2 %g %s2 %g" %
                  (start, tags[3], self.lseg.min, tags[4], self.lseg.max,
                   tags[3], self.useg.min, tags[4], self.useg.max))


class pathElementHintState:
    """Stores the intermediate hint state of a pathElement"""

    def __init__(self) -> None:
        self.s_segs: list[hintSegment] = []
        self.m_segs: list[hintSegment] = []
        self.e_segs: list[hintSegment] = []
        self.mask: list[bool] = []

    def cleanup(self) -> None:
        """Updates and deletes segments according to deleted and replacedBy"""
        for l in (self.s_segs, self.m_segs, self.e_segs):
            l[:] = [x for x in (s.current() for s in l) if not x.deleted]

    def pruneHintSegs(self) -> None:
        """Deletes segments with no assigned hintval"""
        for l in (self.s_segs, self.m_segs, self.e_segs):
            l[:] = [s for s in l if s.hintval is not None]

    def segments(self) -> list:
        return [s for s in self.s_segs + self.m_segs + self.e_segs]

    def segLists(
        self, first=None
    ) -> tuple[HintSegListWithType, HintSegListWithType, HintSegListWithType]:
        if first is None or first == 's':
            return (('s', self.s_segs), ('m', self.m_segs), ('e', self.e_segs))
        elif first == 'm':
            return (('m', self.m_segs), ('s', self.s_segs), ('e', self.e_segs))
        elif first == 'e':
            return (('e', self.e_segs), ('m', self.m_segs), ('s', self.s_segs))
        else:
            raise ValueError('First must be s, m, or e')


class glyphHintState:
    """
    Stores the intermediate hint state (for one dimension) of a glyphData
    object

    peStates: A hash of pathElementHintState objects with the pathElement as
              key
    increasingSegs: Segments with endpoints (in the opposite dimension) greater
                    than their start points
    decreasingSegs: Segments with endpoints (in the opposite dimension) less
                    than their start points
    stemValues: List of stemValue objects in increasing order of position
    mainValues: List of non-overlapping "main" stemValues in decreasing order
                of value
    rejectValues: The set of stemValues - mainValues
    counterHinted: True if the glyph is counter hinted in this dimension
    stems: stemValue stems represented in glyphData 'stem' object format
    weights: weights corresponding to 'stems', for resolving conflicts
    keepHints: If true, keep already defined hints and masks in this dimension
               (XXX only partially implemented)
    hasOverlaps: True when some stemValues overlap
    stemOverlaps: 2d boolean array. Stem n "overlaps" with stem m <=>
                  stemOverlaps[n][m] == True
    ghostCompat: if stem m is a ghost stem, ghostCompat[m] is a boolean
                 array where ghostCompat[m][n] is True <=> n can substitute
                 for m (n has the same location on the relevant side)
    mainMask: glyphData hintmask-like representation of mainValues
    """

    def __init__(self) -> None:
        self.peStates: dict[pathElement, pathElementHintState] = {}
        self.overlapRemoved = None
        self.increasingSegs: list[hintSegment] = []
        self.decreasingSegs: list[hintSegment] = []
        self.stemValues: list[stemValue] = []
        self.mainValues: list[stemValue] = []
        self.rejectValues: list[stemValue] = []
        self.counterHinted = False
        # in sorted glyphData format
        self.stems: Sequence[list[stem]] | None = None
        self.weights: list[float] = []
        self.keepHints = None
        self.hasOverlaps: bool = False
        self.stemOverlaps: list[list[bool]] = []
        self.ghostCompat: list[list[bool] | None] = []
        self.goodMask: list[bool] = []
        self.mainMask: list[bool] = []

    def getPEState(
        self, pe, make: bool = False
    ) -> pathElementHintState | None:
        """
        Returns the pathElementHintState object for pe, allocating the object
        if necessary
        """
        s = self.peStates.get(pe, None)
        if s:
            return s
        if make:
            return self.getPEStateForced(pe)
        else:
            return None

    def getPEStateForced(self, pe) -> pathElementHintState:
        """
        Returns the pathElementHintState object for pe, allocating the object
        if necessary
        """
        return self.peStates.setdefault(pe, pathElementHintState())

    def addSegment(
        self,
        fr,
        to,
        loc,
        pe1: pathElement | None,
        pe2: pathElement | None,
        typ: hintSegment.sType,
        bonus,
        isV: bool,
        mid1,
        mid2,
        desc: str,
    ) -> None:
        """Adds a new segment associated with pathElements pe1 and pe2"""
        if isV:
            pp = ("v", loc, fr, loc, to, desc)
        else:
            pp = ("h", fr, loc, to, loc, desc)
        log.debug("add %sseg %g %g to %g %g %s" % pp)
        if fr > to:
            mn, mx = to, fr
            isInc = False
            lst = self.decreasingSegs
        else:
            mn, mx = fr, to
            isInc = True
            lst = self.increasingSegs

        aSeg = pe2 or pe1
        assert aSeg
        s = hintSegment(loc, mn, mx, aSeg, typ, bonus, isV, isInc, desc)
        # Segments derived from the first point in a path c are typically
        # also added to the previous spline p, with p passed as pe1 and c
        # passed as pe2. Segments derived from the last point in a path
        # are only added to that path, passed as pe1. pe1 is therefore the
        # "end point position" and pe2 the "start point position" except in
        # the special case of MCURVE, which indicates the point was derived
        # from the middle of the spline.
        if pe1:
            if mid1:
                self.getPEStateForced(pe1).m_segs.append(s)
            else:
                self.getPEStateForced(pe1).e_segs.append(s)
        if pe2:
            if mid2:
                self.getPEStateForced(pe2).m_segs.append(s)
            else:
                self.getPEStateForced(pe2).s_segs.append(s)

        bisect.insort(lst, s)

    def compactList(self, l: list[hintSegment]) -> None:
        """
        Compacts overlapping segments with the same location by picking
        one segment to represent the pair, adjusting its values, and
        removing the other segment
        """
        i = 0
        while i < len(l):
            j = i + 1
            while j < len(l) and feq(l[j].loc, l[i].loc):
                if l[i].max >= l[j].min and l[i].min <= l[j].max:
                    if abs(l[i].max - l[i].min) > abs(l[j].max - l[j].min):
                        l[i].min = min(l[i].min, l[j].min)
                        l[i].max = max(l[i].max, l[j].max)
                        l[i].bonus = max(l[i].bonus, l[j].bonus)
                        l[j].replacedBy = l[i]
                        del l[j]
                        continue
                    else:
                        l[j].min = min(l[i].min, l[j].min)
                        l[j].max = max(l[i].max, l[j].max)
                        l[j].bonus = max(l[i].bonus, l[j].bonus)
                        l[i].replacedBy = l[j]
                        del l[i]
                        i -= 1
                        break
                j += 1
            i += 1

    def compactLists(self) -> None:
        """Compacts both segment lists"""
        self.compactList(self.decreasingSegs)
        self.compactList(self.increasingSegs)

    def remExtraBends(self) -> None:
        """
        Delete BEND segment x when there is another segment y:
           1. At the same location
           2. but in the other direction
           2. that is not of type BEND or GHOST and
           3. that overlaps with x and
           4. is at least three times longer
        """
        for hsi in self.increasingSegs:
            lo = bisect.bisect_left(self.decreasingSegs, hsi)
            hi = bisect.bisect_right(self.decreasingSegs, hsi, lo=lo)
            for hsd in self.decreasingSegs[lo:hi]:
                assert hsd.loc == hsi.loc
                if hsd.min < hsi.max and hsd.max > hsi.min:
                    if (hsi.type == hintSegment.sType.BEND and
                            hsd.type != hintSegment.sType.BEND and
                            hsd.type != hintSegment.sType.GHOST and
                            (hsd.max - hsd.min) > (hsi.max - hsi.min) * 3):
                        hsi.deleted = True
                        log.debug("rem seg loc %g from %g to %g" %
                                  (hsi.loc, hsi.min, hsi.max))
                        break
                    elif (hsd.type == hintSegment.sType.BEND and
                          hsi.type != hintSegment.sType.BEND and
                          hsi.type != hintSegment.sType.GHOST and
                          (hsi.max - hsi.min) > (hsd.max - hsd.min) * 3):
                        hsd.deleted = True
                        log.debug("rem seg loc %g from %g to %g" %
                                  (hsd.loc, hsd.min, hsd.max))

    def deleteSegments(self) -> None:
        for s in self.increasingSegs:
            s.deleted = True
        for s in self.decreasingSegs:
            s.deleted = True
        self.increasingSegs = []
        self.decreasingSegs = []
        self.cleanup()

    def cleanup(self) -> None:
        """Runs cleanup on all pathElementHintState objects"""
        self.increasingSegs = [s for s in self.increasingSegs if not s.deleted]
        self.decreasingSegs = [s for s in self.decreasingSegs if not s.deleted]
        for pes in self.peStates.values():
            pes.cleanup()

    def pruneHintSegs(self) -> None:
        """Runs pruneHintSegs on all pathElementHintState objects"""
        for pes in self.peStates.values():
            pes.pruneHintSegs()


class stemLocCandidate:
    strongMultiplier: float = 1.2
    bandMultiplier: float = 2.0
    """
    Information about a candidate location for a stem in a region
    glyph, derived from segments generated for the glyph or,
    occasionally, directly from point locations.
    """

    def __init__(self, loc) -> None:
        self.loc = loc
        self.strong: float = 0
        self.weak: float = 0

    def addScore(self, score: float, strong: bool) -> None:
        if strong:
            self.strong += score
        else:
            self.weak += score

    def weight(self, inBand: bool) -> float:
        return (self.strong * self.strongMultiplier + self.weak) * (
            self.bandMultiplier if inBand else 1
        )

    def isStrong(self) -> bool:
        return self.strong > 0

    def isMixed(self) -> bool:
        return self.strong > 0 and self.weak > 0

    def __eq__(self, other) -> bool:
        return feq(self.strong, other.strong) and feq(self.weak, other.weak)

    def __lt__(self, other) -> bool:
        return self.strong < other.strong or (
            feq(self.strong, other.strong) and self.weak < other.weak
        )


class instanceStemState:
    """
    State for the process of deciding on the lower and upper locations
    for a particular region stem.
    """

    def __init__(self, loc: float, dhinter: AHinter) -> None:
        self.defaultLoc = loc
        self.dhinter = dhinter
        self.candDict: dict[float, stemLocCandidate] = {}
        self.usedSegs: Set[int] = set()
        self.bb: bool | None = None

    def addToLoc(
        self,
        loc: float,
        score: float,
        strong: bool = False,
        bb: bool = False,
        seg: Any = None
    ) -> None:
        if self.bb is not None:
            if self.bb != bb:
                log.info("Mixed bounding-box and "
                         "non-bounding box location data")
        else:
            self.bb = bb
        if seg is not None:
            sid = id(seg)
            if sid in self.usedSegs:
                return
            self.usedSegs.add(sid)
        if self.dhinter.options.roundCoords:
            loc = round(loc)
        if loc in self.candDict:
            sLC = self.candDict[loc]
        else:
            sLC = self.candDict[loc] = stemLocCandidate(loc)
        sLC.addScore(score, strong)

    def bestLocation(self, isBottom: bool) -> Any:
        weights = [(x.weight(self.dhinter.inBand(x.loc, isBottom)), x.loc)
                   for x in self.candDict.values()]
        try:
            m = max(weights)
#            try:
#                next_m = m
#                while abs(next_m[1] - m[1]) < 3.01:
#                    weights.remove(next_m)
#                    next_m = max(weights)
#                if m[0] < next_m[0] * 2:
#                    log.warning("Stem location competition: "
#                                "weight %s vs %s" % (m, next_m))
#            except:
#                pass
            return m[1]
        except ValueError:
            return None


class links:
    """
    Tracks which subpaths need which stem hints and calculates a subpath
    to reduce hint substitution

    cnt: The number of subpaths
    links: A cnt x cnt array of integers modified by mark
           (Values only 0 or 1 but kept as ints for later arithmetic)
    """

    def __init__(self, glyph) -> None:
        l = len(glyph.subpaths)
        if l < 4 or l > 100:
            self.cnt = 0
            return
        self.cnt = l
        self.links = [[0] * l for i in range(l)]

    def logLinks(self) -> None:
        """Prints a log message representing links"""
        if self.cnt == 0:
            return
        log.debug("Links")
        log.debug(' '.join((str(i).rjust(2) for i in range(self.cnt))))
        for j in range(self.cnt):
            log.debug(' '.join((('Y' if self.links[j][i] else ' ').rjust(2)
                                for i in range(self.cnt))))

    def logShort(self, shrt, lab) -> None:
        """Prints a log message representing (1-d) shrt"""
        log.debug(lab)
        log.debug(' '.join((str(i).rjust(2) for i in range(self.cnt))))
        log.debug(' '.join(((str(shrt[i]) if shrt[i] else ' ').rjust(2)
                            for i in range(self.cnt))))

    def mark(self, stemValues: list[stemValue], isV) -> None:
        """
        For each stemValue in hntr, set links[m][n] and links[n][m] to 1
        if one side of a stem is in m and the other is in n
        """
        if self.cnt == 0:
            return
        for sv in stemValues:
            if not sv.lseg or not sv.useg or not sv.lseg.pe or not sv.useg.pe:
                continue
            lseg_pe = sv.lseg.pe()
            useg_pe = sv.useg.pe()
            assert lseg_pe and useg_pe
            lsubp, usubp = lseg_pe.position[0], useg_pe.position[0]
            if lsubp == usubp:
                continue
            sv.show(isV, "mark")
            log.debug(" : %d <=> %d" % (lsubp, usubp))
            self.links[lsubp][usubp] = 1
            self.links[usubp][lsubp] = 1

    def moveIdx(
        self, suborder: list[int], subidxs: list[int], outlinks, idx: int
    ) -> None:
        """
        Move value idx from subidxs to the end of suborder and update
        outlinks to record all links shared with idx
        """
        subidxs.remove(idx)
        suborder.append(idx)
        for i in range(len(outlinks)):
            outlinks[i] += self.links[idx][i]
        self.logShort(outlinks, "Outlinks")

    def shuffle(self) -> list[int] | None:
        """
        Returns suborder list with all subpath indexes in decreasing
        order of links shared with previous subpath. (The first subpath
        being the one with most links overall.)
        """
        if self.cnt == 0:
            return None
        sumlinks = [sum(l) for l in zip(*self.links)]
        outlinks = [0] * self.cnt
        self.logLinks()
        self.logShort(sumlinks, "Sumlinks")
        subidxs = list(range(self.cnt))
        suborder: list[int] = []
        while subidxs:
            # negate s to preserve all-links-equal subpath ordering
            _, bst = max(((sumlinks[s], -s) for s in subidxs))
            self.moveIdx(suborder, subidxs, outlinks, -bst)
            while True:
                try:
                    _, _, bst = max(((outlinks[s], sumlinks[s], -s)
                                     for s in subidxs if outlinks[s] > 0))
                except ValueError:
                    break
                self.moveIdx(suborder, subidxs, outlinks, -bst)
        return suborder



# Copyright 2022 Adobe. All rights reserved.

import argparse
from copy import deepcopy
import logging
import plistlib
import re
import sys

from fontTools.ttLib import TTFont
from fontTools.cffLib import FontDict, FDArrayIndex, PrivateDict, FDSelect
from fontTools.designspaceLib import DesignSpaceDocument
from afdko.fdkutils import (
    run_shell_command,
    get_temp_file_path,
    validate_path,
)

logger = logging.getLogger(__name__)


class ShellCommandError(Exception):
    pass


def replaceCFF(otfPath: str, tempFilePath: str) -> None:
    if not run_shell_command(['sfntedit', '-a',
                              f'CFF ={tempFilePath}', otfPath]):
        raise ShellCommandError


def get_options(args: list[str] | None) -> argparse.Namespace:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-m',
        '--metadata',
        dest='metaPath',
        type=validate_path,
        help='path to hinting metadata file',
        required=True
    )

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '-d',
        '--designspace',
        dest='dsPath',
        type=validate_path,
        help='path to design space file',
    )
    group.add_argument(
        '-f',
        '--fontpath',
        dest='fPath',
        type=validate_path,
        help='path to individual OTF font to modify',
    )
    options = parser.parse_args(args)

    return options


class Dictspec():
    name: str
    re: list[re.Pattern]
    BluePairs: list[int]
    OtherPairs: list[int]
    SIV: list[int]
    SIH: list[int]
    foundGlyph: bool

    pass


def getDictmap(options: argparse.Namespace) -> list[Dictspec]:
    with open(options.metaPath, "rb") as dictmapfile:
        rawdictmap = plistlib.load(dictmapfile)

    dictmap = []
    for name, d in rawdictmap.items():
        dct = Dictspec()
        dct.name = name
        dct.BluePairs = d.get("blue_pairs", [])
        dct.OtherPairs = d.get("other_blue_pairs", [])
        dct.SIV = d.get("stem_indexes_v", [])
        dct.SIH = d.get("stem_indexes_h", [])
        dct.re = []
        for r in d.get("regex", []):
            dct.re.append(re.compile(r))
        dct.foundGlyph = False
        dictmap.append(dct)

    return dictmap


def remapDicts(fpath: str, dictmap: list[Dictspec]) -> None:
    f = TTFont(fpath)
    if 'CFF ' not in f:
        logger.error("No CFF table in %s: will not modify" % fpath)
        return

    cff = f['CFF ']
    top = cff.cff[0]
    if hasattr(top, "FDArray"):
        logger.error("CFF table in %s already has " % fpath +
                     "%d private dicts: will not modify" % len(top.FDArray))
        return

    hasExtra = False

    # Create a new FDSelect object for the font and map the glyphs to
    # dictionary indexes according to the regular expressions in
    # Dictspec.re
    fdselect = top.FDSelect = FDSelect()
    for gn in cff.getGlyphOrder():
        done = False
        for i, dictspec in enumerate(dictmap):
            for r in dictspec.re:
                if r.search(gn):
                    fdselect.append(i)
                    dictspec.foundGlyph = True
                    done = True
                    break
            if done:
                break
        if not done:
            fdselect.append(i + 1)
            hasExtra = True

    # Rebuild the dictionary indices and the dictmap itself to remove
    # any dictionaries that didn't contain any glyphs, so that we don't
    # wind up with empty dictionaries later.
    j = 0
    indexMap = []
    newdictmap: list[Dictspec] = []
    for dictspec in dictmap:
        if dictspec.foundGlyph:
            newdictmap.append(dictspec)
            indexMap.append(j)
            j += 1
        else:
            logger.warning("No glyphs found for entry %s, skipping" %
                           dictspec.name)
            indexMap.append(-1)
    if hasExtra:
        indexMap.append(j)

    for i, v in enumerate(fdselect):
        assert indexMap[v] != -1
        fdselect[i] = indexMap[v]

    # Create new FDArray and top-level private dictionaries, keeping
    # a reference to the original private dictionary in origPriv
    fdarray = top.FDArray = FDArrayIndex()
    origPriv = top.Private
    # Make sure info doesn't leak though in non-FDArray private dict
    top.Private = PrivateDict()

    # For each dictionary containing glyphs, make a copy of the original
    # private dictionary and then modify the Other/Family/Blues, StemSnap,
    # and Std?W values according to the dictspec. (Note that by assigning
    # to blah[:] the new values wind up in the lists inside the dictionary
    # copy.)
    #
    # Then add that dictionary to a new FontDict and append the later
    # to the fdarray.
    for dictspec in newdictmap:
        # Make a new dictonary starting with the old private dictionary
        # values
        npr = deepcopy(origPriv)
        for bvals, fvals, indexes in [(getattr(npr, "BlueValues", None),
                                       getattr(npr, "FamilyBlues", None),
                                       dictspec.BluePairs),
                                      (getattr(npr, "OtherBlues", None),
                                       getattr(npr, "FamilyOtherBlues", None),
                                       dictspec.OtherPairs)]:
            nbvals = []
            nfvals = []
            for pi in sorted(indexes):
                if bvals:
                    nbvals.extend([bvals[2 * pi], bvals[2 * pi + 1]])
                if fvals:
                    nfvals.extend([fvals[2 * pi], fvals[2 * pi + 1]])
            if bvals:
                bvals[:] = nbvals
            if fvals:
                fvals[:] = nfvals
        snapvs = getattr(npr, "StemSnapV", None)
        if snapvs is not None:
            npr.StemSnapV[:] = [snapvs[i] for i in dictspec.SIV]
            npr.StdVW = npr.StemSnapV[0]
        elif getattr(npr, "StdVW", None) is None:
            logger.error("No StemSnapV or StdVW in combined dict")

        snaphs = getattr(npr, "StemSnapH", None)
        if snaphs is not None:
            npr.StemSnapH[:] = [snaphs[i] for i in dictspec.SIH]
            npr.StdHW = npr.StemSnapH[0]
        elif getattr(npr, "StdHW", None) is None:
            logger.error("No StemSnapH or StdHW in combined dict")

        fontDict = FontDict()
        fontDict.setCFF2(True)
        fontDict.Private = npr
        fdarray.append(fontDict)

    # Add a dictionary for any remaining glyphs with zones far outside of
    # the bounds of existing glyphs. (Leave the stem values.)
    if hasExtra:
        bbox = top.FontBBox
        font_max = bbox[3]
        font_min = bbox[1]
        npr = deepcopy(fdarray[0].Private)  # To get reasonable stem values
        npr.BlueValues[:] = [font_min - 100, font_min - 85,
                             font_max + 85, font_max + 100]
        if hasattr(npr, "FamilyBlues"):
            npr.FamilyBlues[:] = [font_min - 100, font_min - 85,
                                  font_max + 85, font_max + 100]
        if hasattr(npr, "OtherBlues"):
            npr.OtherBlues[:] = []
        if hasattr(npr, "FamilyOtherBlues"):
            npr.FamilyOtherBlues[:] = []
        fontDict = FontDict()
        fontDict.setCFF2(True)
        fontDict.Private = npr
        fdarray.append(fontDict)

    numdicts = len(newdictmap) + 1 if hasExtra else 0

    logger.warning("Updating %s with %d dictionaries" % (fpath, numdicts))
    cfftpath = get_temp_file_path()
    cfftfile = open(cfftpath, "w+b")
    cff.cff.compile(cfftfile, f, False)
    cfftfile.close()
    f.close()

    replaceCFF(fpath, cfftpath)


def main(args: list[str] | None = None) -> int:
    options = get_options(args)

    dictmap = getDictmap(options)

    if options.dsPath is not None:
        ds = DesignSpaceDocument.fromfile(options.dsPath)
        otf_paths = [s.path.replace('.ufo', '.otf') for s in ds.sources]
    else:
        otf_paths = [options.fPath]

    for p in otf_paths:
        remapDicts(p, dictmap)

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
AFDKO Unified Command Invoker

This module provides the command registry, help system, and dispatch logic
for the unified 'afdko' command.
"""

import sys
from typing import NoReturn


# Complete command registry with abbreviations
# Format: name -> (module:function, description, category)
# Categories: 'primary', 'secondary', 'plot'

ALL_COMMANDS = {
    # PRIMARY COMMANDS - Most commonly used
    'tx': ('afdko._internal:tx', 'Font converter and analyzer',
           'primary'),
    'spot': ('afdko._internal:spot', 'SFNT font inspector', 'primary'),
    'sfntedit': ('afdko._internal:sfntedit', 'SFNT table editor',
                 'primary'),
    'se': ('afdko._internal:sfntedit', None, 'primary'),  # abbrev
    'mergefonts': ('afdko._internal:mergefonts', 'Merge font files',
                   'primary'),
    'mf': ('afdko._internal:mergefonts', None, 'primary'),  # abbrev
    'rotatefont': ('afdko._internal:rotatefont', 'Rotate font glyphs',
                   'primary'),
    'rf': ('afdko._internal:rotatefont', None, 'primary'),  # abbrev
    'makeotf': ('afdko.makeotf:main', 'Build OpenType font', 'primary'),
    'mo': ('afdko.makeotf:main', None, 'primary'),  # abbrev
    'buildcff2vf': ('afdko.buildcff2vf:main', 'Build CFF2 variable font',
                    'primary'),
    'bvf': ('afdko.buildcff2vf:main', None, 'primary'),  # abbrev
    'buildmasterotfs': ('afdko.buildmasterotfs:main',
                        'Build master OTF fonts', 'primary'),
    'bmo': ('afdko.buildmasterotfs:main', None, 'primary'),  # abbrev
    'makeinstancesufo': ('afdko.makeinstancesufo:main',
                         'Generate UFO instances', 'primary'),
    'miu': ('afdko.makeinstancesufo:main', None, 'primary'),  # abbrev
    'checkoutlinesufo': ('afdko.checkoutlinesufo:main',
                         'Check UFO outlines', 'primary'),
    'cou': ('afdko.checkoutlinesufo:main', None, 'primary'),  # abbrev
    'otfautohint': ('afdko.otfautohint.__main__:main', 'Auto-hint fonts',
                    'primary'),
    'autohint': ('afdko.otfautohint.__main__:main', None,
                 'primary'),  # abbrev
    'ah': ('afdko.otfautohint.__main__:main', None, 'primary'),  # abbrev
    'otfstemhist': ('afdko.otfautohint.__main__:stemhist',
                    'Generate stem histogram', 'primary'),
    'stemhist': ('afdko.otfautohint.__main__:stemhist', None,
                 'primary'),  # abbrev
    'sh': ('afdko.otfautohint.__main__:stemhist', None,
           'primary'),  # abbrev
    'addfeatures': ('afdko._internal:addfeatures',
                    'Feature file compiler', 'primary'),
    'af': ('afdko._internal:addfeatures', None, 'primary'),  # abbrev

    # SECONDARY COMMANDS - Less commonly used
    'type1': ('afdko._internal:type1', 'Type 1 font tool', 'secondary'),
    't1': ('afdko._internal:type1', None, 'secondary'),  # abbrev
    'detype1': ('afdko._internal:detype1', 'Type 1 font decompiler',
                'secondary'),
    'dt1': ('afdko._internal:detype1', None, 'secondary'),  # abbrev
    'sfntdiff': ('afdko._internal:sfntdiff', 'Compare SFNT fonts',
                 'secondary'),
    'comparefamily': ('afdko.comparefamily:main', 'Compare font family',
                      'secondary'),
    'cf': ('afdko.comparefamily:main', None, 'secondary'),  # abbrev
    'otc2otf': ('afdko.otc2otf:main', 'Extract OTFs from OTC',
                'secondary'),
    'otf2otc': ('afdko.otf2otc:main', 'Combine OTFs into OTC',
                'secondary'),
    'otf2ttf': ('afdko.otf2ttf:main', 'Convert OTF to TTF', 'secondary'),
    'ttfcomponentizer': ('afdko.ttfcomponentizer:main',
                         'Add TTF components', 'secondary'),
    'ttfdecomponentizer': ('afdko.ttfdecomponentizer:main',
                           'Remove TTF components', 'secondary'),
    'ttxn': ('afdko.ttxn:main', 'TTX wrapper', 'secondary'),

    # CONFIG COMMANDS - Configuration and setup
    'completion': ('afdko.completion:main', 'Generate shell completion script',
                   'config'),

    # PLOT COMMANDS - Proofing tools
    'charplot': ('afdko.proofpdf:charplot', 'Generate character proof',
                 'plot'),
    'digiplot': ('afdko.proofpdf:digiplot', 'Generate digitization proof',
                 'plot'),
    'fontplot': ('afdko.proofpdf:fontplot', 'Generate font proof', 'plot'),
    'fontplot2': ('afdko.proofpdf:fontplot2', 'Generate font proof (v2)',
                  'plot'),
    'fontsetplot': ('afdko.proofpdf:fontsetplot',
                    'Generate font set proof', 'plot'),
    'hintplot': ('afdko.proofpdf:hintplot', 'Generate hint proof', 'plot'),
    'waterfallplot': ('afdko.proofpdf:waterfallplot',
                      'Generate waterfall proof', 'plot'),
}


def print_category_commands(category: str) -> None:
    """Print commands for a specific category."""
    seen = set()
    for name, (target, desc, cat) in ALL_COMMANDS.items():
        if cat == category and desc is not None and target not in seen:
            # Find abbreviations for this command
            abbrevs = [k for k, (t, d, c) in ALL_COMMANDS.items()
                       if t == target and d is None and c == category]
            if abbrevs:
                abbrev_str = f" ({', '.join(abbrevs)})"
            else:
                abbrev_str = ""
            print(f"  {name:20} {desc}{abbrev_str}")
            seen.add(target)


def print_help(show_category: str = 'primary') -> None:
    """Print help message.

    Args:
        show_category: 'primary', 'secondary', 'plot', or 'all'
    """
    print("Usage: afdko <command> [options]")
    print("   or: afdko -h <command>  (get help for specific command)")
    print()
    print("AFDKO (Adobe Font Development Kit for OpenType) is "
          "Adobe's open source")
    print("toolkit for building, editing, validating, and proofing "
          "OpenType fonts")
    print("with emphasis on PostScript-lineage CFF and CFF2 output. "
          "It includes over")
    print("25 specialized tools for tasks ranging from compiling fonts "
          "with OpenType")
    print("layout features (makeotf), auto-hinting (otfautohint), and "
          "format conversion")
    print("(tx, otf2ttf), to quality validation (checkoutlinesufo, "
          "comparefamily) and")
    print("generating detailed proofs (spot, charplot).")
    print()

    if show_category in ('primary', 'all'):
        print("Primary Commands:")
        print_category_commands('primary')
        print()

    # Config commands shown with primary (distinctive display)
    if show_category in ('primary', 'all'):
        print("Configuration:")
        print_category_commands('config')
        print()

    if show_category in ('secondary', 'all'):
        print("Secondary Commands:")
        print_category_commands('secondary')
        print()

    if show_category in ('plot', 'all'):
        print("Proofing Commands:")
        print_category_commands('plot')
        print()

    if show_category == 'primary':
        print("Additional commands:")
        print("  afdko -s         List secondary commands")
        print("  afdko -p         List proofing commands")
        print("  afdko -a         List all commands")
        print()

    print("Run 'afdko <command> -h' or 'afdko -h <command>' for "
          "command-specific help.")


def dispatch_command(cmd: str) -> NoReturn:
    """Dispatch to the appropriate command implementation."""
    if cmd not in ALL_COMMANDS:
        print(f"Error: Unknown command '{cmd}'", file=sys.stderr)
        print("Run 'afdko --help' for usage.", file=sys.stderr)
        sys.exit(1)

    target, _, _ = ALL_COMMANDS[cmd]  # Unpack 3-tuple
    module_name, func_name = target.split(':')

    # Import and call the function
    try:
        module = __import__(module_name, fromlist=[func_name])
        func = getattr(module, func_name)
    except (ImportError, AttributeError) as e:
        print(f"Error loading command '{cmd}': {e}", file=sys.stderr)
        sys.exit(1)

    # Adjust sys.argv to look like the command was called directly
    # "afdko makeotf -f font.pfa" becomes "makeotf -f font.pfa"
    sys.argv = sys.argv[1:]

    # Call the function - it will handle sys.exit() internally
    result = func()

    # Handle return value (some commands return exit code)
    if result is None:
        sys.exit(0)
    elif isinstance(result, int):
        sys.exit(result)
    else:
        sys.exit(0)


def main() -> NoReturn:
    """Main entry point for the unified afdko command."""
    # No subcommand provided
    if len(sys.argv) < 2:
        print_help('primary')
        sys.exit(1)

    subcmd = sys.argv[1]

    # Top-level help flags
    if subcmd in ('-s', '--secondary'):
        print_help('secondary')
        sys.exit(0)
    elif subcmd in ('-p', '--plot'):
        print_help('plot')
        sys.exit(0)
    elif subcmd in ('-a', '--all'):
        print_help('all')
        sys.exit(0)

    # Help requested
    if subcmd in ('-h', '--help', 'help'):
        # Check for command name after -h
        if len(sys.argv) > 2:
            arg = sys.argv[2]
            # afdko -h <command> -> afdko <command> -h
            # Check if it's a valid command
            if arg in ALL_COMMANDS:
                sys.argv = ['afdko', arg, '-h']
                dispatch_command(arg)
            else:
                print(f"Error: Unknown command '{arg}'", file=sys.stderr)
                print("Run 'afdko --help' for usage.", file=sys.stderr)
                sys.exit(1)
        else:
            print_help('primary')
            sys.exit(0)

    # Dispatch to command
    dispatch_command(subcmd)


if __name__ == '__main__':
    main()



# Copyright 2016 Adobe. All rights reserved.

# Methods:

# Parse args. If glyphlist is from file, read in entire file as single string,
# and remove all white space, then parse out glyph-names and GID's.

import logging
import os
import time
from threading import Thread
from multiprocessing import Pool, Manager, current_process
from typing import Any, Iterator, NamedTuple

from .otfFont import CFFFontData
from .ufoFont import UFOFontData
from .report import Report, GlyphReport
from .hinter import glyphHinter
from .logging import log_receiver
from .fdTools import FDDictManager
from .glyphData import glyphData

from . import (get_font_format, FontParseError)

# Import from __main__ for type annotations
# (avoid circular import by using TYPE_CHECKING)
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .__main__ import HintOptions, ReportOptions

log = logging.getLogger(__name__)


class ACOptions(object):
    def __init__(self) -> None:
        self.inputPaths: list[str] = []
        self.outputPaths: list[str] = []
        self.referenceFont = None
        self.referenceOutputPath = None
        self.glyphList: list[str] = []
        # True when contents of glyphList were specified directly by the user
        self.explicitGlyphs = False
        self.nameAliases: dict[str, str] = {}
        self.excludeGlyphList = False
        self.overlapList: list[str] = []
        self.overlapForcing: bool | None = None
        self.looseOverlapMapping = False
        self.hintAll = False
        self.readHints = True
        self.allowChanges = False
        self.noFlex = False
        self.noHintSub = False
        self.allowNoBlues = False
        self.fontinfoPath = None
        self.ignoreFontinfo = False
        self.logOnly = False
        self.removeConflicts = True
        # Copy of parse_args verbose for child processes
        self.verbose = 0
        self.printFDDictList = False
        self.printAllFDDict = False
        self.roundCoords = True
        self.writeToDefaultLayer = False
        # If this number of segments is exceeded in a dimension, don't hint
        # (Only applies when explicitGlyphs is False)
        self.maxSegments = 100
        self.font_format = None
        self.report_zones = False
        self.report_stems = False
        self.report_all_stems = False
        self.process_count = None
        # False in these dictionaries indicates that there should be no
        # warning if the glyph is missing
        self.hCounterGlyphs = {'element': False, 'equivalence': False,
                               'notelement': False, 'divide': False}
        self.vCounterGlyphs = {'m': False, 'M': False, 'T': False,
                               'ellipsis': False}
        self.upperSpecials = {'questiondown': False, 'exclamdown': False,
                              'semicolon': False}
        self.lowerSpecials = {'question': False, 'exclam': False,
                              'colon': False}
        self.noBlues = {'at': False, 'bullet': False, 'copyright': False,
                        'currency': False, 'registered': False}
#        self.newHintsOnMoveto = {'percent': False, 'perthousand': False}

    def __str__(self) -> str:
        # used only when debugging.
        import inspect
        data = []
        methodList = inspect.getmembers(self)
        for fname, fvalue in methodList:
            if fname[0] == "_":
                continue
            data.append(str((fname, fvalue)))
        data.append("")
        return os.linesep.join(data)

    def justReporting(self) -> bool:
        return self.report_zones or self.report_stems


def getGlyphNames(
    glyphSpec: str,
    fontGlyphList: list[str],
    fDesc: str
) -> list[str] | None:
    # If the "range" is actually in the font, just ignore its apparent
    # range-ness
    if glyphSpec.isnumeric():
        GID = int(glyphSpec)
        if GID >= len(fontGlyphList):
            log.warning("glyph ID <%s> from glyph selection list option "
                        "exceeds number of glyphs in font <%s>.",
                        glyphSpec, fDesc)
            return None
        return [fontGlyphList[GID]]
    elif glyphSpec in fontGlyphList:
        return [glyphSpec]

    rangeList = glyphSpec.split("-")
    if len(rangeList) == 1:
        # Not numeric, not name, not range
        log.warning("glyph name <%s> from glyph selection list option is "
                    "not in font <%s>", glyphSpec, fDesc)
        return None
    elif len(rangeList) > 2:
        log.warning("Too many hyphens in glyph selection range <%s>",
                    glyphSpec)
        return None

    glyphName1, glyphName2 = rangeList
    gidList = []

    for r in rangeList:
        if r.isnumeric():
            GID = int(r)
            if GID >= len(fontGlyphList):
                log.warning("glyph name <%s> in range %s from glyph "
                            "selection list option is not in font <%s>.",
                            r, glyphSpec, fDesc)
                return None
        else:
            try:
                GID = fontGlyphList.index(r)
            except ValueError:
                log.warning("glyph name <%s> in range %s from glyph "
                            "selection list option is not in font <%s>.",
                            r, glyphSpec, fDesc)
                return None
        gidList.append(GID)

    glyphNameList = []
    for i in range(gidList[0], gidList[1] + 1):
        glyphNameList.append(fontGlyphList[i])

    return glyphNameList


def filterGlyphList(
    options: "HintOptions | ReportOptions",
    fontGlyphList: list[str],
    fDesc: str
) -> list[str]:
    """
    Returns the list of glyphs which are in the intersection of the argument
    list and the glyphs in the font.
    """
    if not options.glyphList:
        glyphList = fontGlyphList
    else:
        # expand ranges:
        glyphList = []
        for glyphSpec in options.glyphList:
            glyphNames = getGlyphNames(glyphSpec, fontGlyphList, fDesc)
            if glyphNames is not None:
                glyphList.extend(glyphNames)
        if options.excludeGlyphList:
            glyphList = [n for n in fontGlyphList if n not in glyphList]
    return glyphList


def get_glyph(
    options: "HintOptions | ReportOptions",
    font: UFOFontData | CFFFontData,
    name: str
) -> glyphData | None:
    try:
        gl = font.convertToGlyphData(name, options.readHints,  # stems
                                     options.readHints,  # flex
                                     options.roundCoords,
                                     options.hintAll)
        if gl is None or gl.isEmpty():
            log.info("Skipping glyph %s: no data from convertToGlyphData" %
                     name)
            return None
        return gl
    except KeyError:
        # Source fonts may be sparse, e.g. be a subset of the
        # reference font.
        return None


class FontInstance(NamedTuple):
    font: Any
    inpath: str
    outpath: str | None


def setUniqueDescs(fontInstances: list[FontInstance]) -> None:
    descs = [f.font.getPSName() for f in fontInstances]
    if None in descs or '' in descs or len(descs) != len(set(descs)):
        if len(fontInstances) == 1:
            # Don't add a substantive instance path when hinting one font
            fontInstances[0].font.desc = ''
            return
        descs = [f.inpath for f in fontInstances]
        try:
            prefix = os.path.commonpath(descs)
            descs = [os.path.relpath(p, prefix) for p in descs]
            while True:
                dirname = [os.path.dirname(p) for p in descs]
                if len(set(os.path.basename(p) for p in descs)) == 1:
                    descs = dirname
                else:
                    break
        except ValueError:
            pass
    for f, d in zip(fontInstances, descs):
        f.font.desc = d


class fontWrapper:
    """
    Stores references to one or more related instance font objects.
    Extracts glyphs from those objects by name, hints them, and
    stores the result back those objects. Optionally saves the
    modified glyphs in corresponding output font files.
    """
    def __init__(
        self,
        options: "HintOptions | ReportOptions",
        fil: list[FontInstance]
    ) -> None:
        self.options = options
        self.fontInstances = fil
        setUniqueDescs(fil)
        if len(fil) > 1:
            self.isVF = False
        else:
            self.isVF = (hasattr(fil[0].font, 'ttFont') and
                         'fvar' in fil[0].font.ttFont)
        self.reportOnly = options.justReporting()
        assert not self.reportOnly or (not self.isVF and len(fil) == 1)
        self.notFound = 0
        self.glyphNameList = filterGlyphList(options,
                                             fil[0].font.getGlyphList(),
                                             fil[0].font.desc)
        if not self.glyphNameList:
            raise FontParseError("Selected glyph list is empty for " +
                                 "font <%s>." % fil[0].font.desc)
        self.dictManager = FDDictManager(options, fil, self.glyphNameList,
                                         self.isVF)

    def numGlyphs(self) -> int:
        return len(self.glyphNameList)

    def hintStatus(self, name: str, hintedGlyphTuple: Any) -> bool:
        an = self.options.nameAliases.get(name, name)
        if hintedGlyphTuple is None:
            log.warning("%s: Could not hint!", an)
            return False
        hs = [g.hasHints(both=True) for g in hintedGlyphTuple if g is not None]
        if False in hs:
            if len(hintedGlyphTuple) == 1:
                log.info("%s: No hints added!", an)
                return False
            elif True in hs:
                log.info("%s: Hints only added to some instances!", an)
                return True
            else:
                log.info("%s: No hints added to any instances!", an)
                return False
        return True

    class glyphiter:
        def __init__(self, parent: "fontWrapper") -> None:
            self.fw = parent
            self.gnIter = parent.glyphNameList.__iter__()
            self.notFound = 0

        def __next__(self) -> tuple[str, tuple[glyphData], tuple[int, int]]:
            # gnIter's StopIteration exception stops this iteration
            vsindex = 0
            stillLooking = True
            while stillLooking:
                stillLooking = False
                name = self.gnIter.__next__()
                if self.fw.reportOnly and name == '.notdef':
                    stillLooking = True
                    continue
                if self.fw.isVF:
                    t = self.fw.fontInstances[0].font.get_vf_glyphs(name)
                    gt, vsindex = t

                else:
                    gt = tuple((get_glyph(self.fw.options, f.font, name)
                                for f in self.fw.fontInstances))
                if True not in (g is not None for g in gt):
                    self.notFound += 1
                    stillLooking = True
            self.fw.notFound = self.notFound
            return name, gt, self.fw.dictManager.getRecKey(name, vsindex)

    def __iter__(
        self
    ) -> Iterator[tuple[str, tuple[glyphData], tuple[int, int]]]:
        return self.glyphiter(self)  # type: ignore

    def hint(self) -> bool:
        hintedAny = False

        report = Report() if self.reportOnly else None

        pcount = self.options.process_count
        if pcount is None:
            pcount = os.cpu_count()
        assert pcount is not None
        if pcount < 0:
            cpu_count = os.cpu_count()
            assert cpu_count is not None
            pcount = cpu_count - pcount
            if pcount < 0:
                pcount = 1
        if pcount > self.numGlyphs():
            pcount = self.numGlyphs()

        if pcount > 1 and current_process().daemon:
            pcount = 1
            if self.options.process_count is not None:
                log.warning("Program was already spawned by a different " +
                            "python process: running as single process")

        pool = None
        logThread = None
        gmap: Iterator | None = None
        try:
            dictRecord = self.dictManager.getDictRecord()
            if pcount == 1:
                glyphHinter.initialize(self.options, dictRecord)
                gmap = map(glyphHinter.hint, self)
            else:
                # set_start_method('spawn')
                manager = Manager()
                logQueue = manager.Queue(-1)
                logThread = Thread(target=log_receiver, args=(logQueue,))
                logThread.start()
                pool = Pool(pcount, initializer=glyphHinter.initialize,
                            initargs=(self.options, dictRecord,
                                      logQueue))
                if report is not None:
                    # Retain glyph ordering for reporting purposes
                    gmap = pool.imap(glyphHinter.hint, self)
                else:
                    gmap = pool.imap_unordered(glyphHinter.hint, self)

            for name, r in gmap:
                if isinstance(r, GlyphReport):
                    if report is not None:
                        report.glyphs[name] = r
                else:
                    hasHints = self.hintStatus(name, r)
                    if r is None:
                        r = [None] * len(self.fontInstances)
                    if not self.options.logOnly:
                        if hasHints:
                            hintedAny = True
                        font = self.fontInstances[0].font
                        for i, new_glyph in enumerate(r):
                            if i > 0 and not self.isVF:
                                font = self.fontInstances[i].font
                            font.updateFromGlyph(new_glyph, name)
                        if self.isVF:
                            font.merge_hinted_glyphs(name)

            if self.notFound:
                log.info("Skipped %s of %s glyphs.", self.notFound,
                         self.numGlyphs())

            if report is not None:
                outpath = self.fontInstances[0].outpath
                assert outpath is not None
                report.save(outpath, self.options)
            elif not hintedAny:
                log.info("No glyphs were hinted.")

            if pool is not None:
                pool.close()
                pool.join()
                logQueue.put(None)
                assert logThread is not None
                logThread.join()
        finally:
            if pool is not None:
                pool.terminate()
                pool.join()
            if logThread is not None:
                logQueue.put(None)
                logThread.join()

        return hintedAny

    def save(self) -> None:
        for f in self.fontInstances:
            log.info("Saving font file %s with new hints..." % f.outpath)
            f.font.save(f.outpath)

    def close(self) -> None:
        for f in self.fontInstances:
            log.info("Closing font file %s without saving." % f.outpath)
            f.font.close()


def openFont(
    path: str,
    options: "HintOptions | ReportOptions"
) -> UFOFontData | CFFFontData:
    font_format = get_font_format(path)
    if font_format is None:
        raise FontParseError(f"{path} is not a supported font format")

    if font_format == "UFO":
        return UFOFontData(path, options.logOnly, options.writeToDefaultLayer)
    else:
        return CFFFontData(path, font_format)


def get_outpath(
    options: "HintOptions | ReportOptions",
    font_path: str,
    i: int | str
) -> str:
    if i == 'r':
        if options.referenceOutputPath is not None:
            outpath = options.referenceOutputPath
        else:
            outpath = font_path
    else:
        assert isinstance(i, int)
        if options.outputPaths is not None and i < len(options.outputPaths):
            outpath = options.outputPaths[i]
        else:
            outpath = font_path
    return outpath


def hintFiles(options: "HintOptions | ReportOptions") -> None:
    fontInstances = []
    # If there is a reference font, prepend it to font paths.
    # It must be the first font in the list, code below assumes that.
    if options.referenceFont:
        font = openFont(options.referenceFont, options)
        if hasattr(font, 'ttFont'):
            assert 'fvar' not in font.ttFont, ("Can't use a CFF2 VF font as a "
                                               "default font in a set of MM "
                                               "fonts.")
        fontInstances.append(FontInstance(font,
                             os.path.abspath(options.referenceFont),
                             get_outpath(options, options.referenceFont, 'r')))

    # Open the rest of the fonts and handle output paths.
    for i, path in enumerate(options.inputPaths):
        fontInstances.append(FontInstance(openFont(path, options),
                                          os.path.abspath(path),
                                          get_outpath(options, path, i)))

    noFlex = options.noFlex
    if fontInstances and options.referenceFont:
        log.info("Hinting fonts with reference %s. Start time: %s.",
                 fontInstances[0].font.desc, time.asctime())
        if fontInstances[0].font.isCID():
            options.noFlex = True
        fw = fontWrapper(options, fontInstances)
        if fw.hint():
            fw.save()
        else:
            fw.close()
        log.info("Done hinting fonts with reference %s. End time: %s.",
                 fontInstances[0].font.desc, time.asctime())
    else:
        for fi in fontInstances:
            log.info("Hinting font %s. Start time: %s.", fi.font.desc,
                     time.asctime())
            if fi[0].isCID():
                options.noFlex = True  # Flex hinting in CJK fonts does
                # bad things. For CFF fonts, being a CID font is a good
                # indicator of being CJK.
            else:
                options.noFlex = noFlex
            fw = fontWrapper(options, [fi])
            if fw.hint():
                fw.save()
            else:
                fw.close()
            log.info("Done hinting font %s. End time: %s.", fi.font.desc,
                     time.asctime())
